__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import re
import types

import http
import threadserver as server

from pyy.web import httperror

class httpserver(object):
  compress = True
  rewrite = []
  sites   = []
  uri     = []
  port    = 8080

  def __init__(self):
    if not self.sites:
      self.sites = [True, self.uri]


  def find_handler(self, uri, config, *args):
    handler = False
    for r, c in reversed(zip(config[0::2], config[1::2])):
      h = False
      # unconditionaly matches
      if r is True: h = c
      # matches None or empty string
      elif r is None:
        if not uri: h = c
      # matches that error code
      elif isinstance(r, int):
        if isinstance(uri, httperror) and uri.args[0] == r:
          h = c
      # matches all errors
      elif r is httperror:
        if isinstance(uri, httperror):  h = c
      # matches regex string
      elif isinstance(r, str):
        if isinstance(uri, str):
          m = re.match(r, uri)
          if m:
            args += m.groups()
            uri = uri[:m.span()[1]]
            h = c
      # didn't match anything
      if h is not False:
        handler = h
        break
    return handler, uri, args


  def handle(self, conn, req, res):
    handler = None
    args = ()

    # rewrite uris
    for r,s in self.rewrite:
      req.uri = re.sub(r, s, req.uri)

    # find a handler
    handler, uri, args = self.find_handler(req.host, self.sites)
    uri = req.uri
    while 1:
      if isinstance(handler, str):
        handler, uri, args = self.find_handler(handler, self.sites)

      if isinstance(handler, list):
        handler, uri, args = self.find_handler(uri, handler)

      elif isinstance(handler, httperror):
        raise handler

      elif isinstance(handler, (types.FunctionType, types.MethodType)):
        handler = handler(*args)
        args = ()

      elif hasattr(handler, 'handle'):
        return handler.handle(conn, req, res, *args)

      elif handler is False:  raise httperror(404)
      elif handler is None:   return False
      else:                   raise ValueError(handler)

  def handle_error(self, conn, req, res, status, *errors):
    if status == 500:
      import traceback
      res.body = traceback.format_exc()

    elif status in (301, 302, 303):
      if errors:
        res.headers['Location'] = errors[0]

    handler, uri, args = self.find_handler(req and req.host, self.sites)
    uri = httperror(status, *errors)
    try:
      while 1:
        if isinstance(handler, list):
          handler, uri, args = self.find_handler(uri, handler)

        elif isinstance(handler, httperror):
          raise handler

        elif isinstance(handler, (types.FunctionType, types.MethodType)):
          handler = handler(*args)
          args = ()

        elif hasattr(handler, 'handle_error'):
          return handler.handle_error(conn, req, res, status, *args)

        # why is this two different cases?
        elif handler is None:
          if status == 500:
            import traceback
            traceback.print_exc()
          return False
        elif handler is False:
          if status == 500:
            import traceback
            traceback.print_exc()
          return
        else: raise ValueError(handler)

    except httperror, e:
      if e.args == (status,)+errors:
        return self.handle_error(conn, req, res, 500, 'infinite recursion in error handler')
      return self.handle_error(conn, req, res, e.args[0], *e.args[1:])


  def run(self, *args):
    s = server.server()
    def handle(server, conn):
      return http.httphandler(conn, self)

    s.listen(('',self.port), handle)
    s.run()




__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

# http://www.w3.org/Protocols/rfc2616/rfc2616.html
import warnings
import time
import datetime
import socket

from pyy.web import httprequest, httpresponse, httperror

def httptime(t=None):
  fmt = '%a, %d %b %Y %H:%M:%S GMT' # http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.3.1
  if isinstance(t, str):
    return datetime.datetime.strptime(t, fmt)
  if isinstance(t, float):
    return time.strftime(fmt, t)
  return datetime.datetime.utcnow().strftime(fmt)

CRLF = '\r\n'

class endconnection(EOFError): pass

class httphandler(object):
  '''
  httphandler class

  httphandler(conn, handler)
    conn - an object that supports blocking
              .read([n])
              .unread(data)
              .write(data)
              .close()


    handler - http server confoguration object.
            will document later



    this class reads data from the connection,
    processes HTTP requests,
    lets handler handle them,
    and writes the results back over HTTP.

    Supports HTTP/1.1 and pipelining

  '''
  id = 0
  def __init__(self, conn, handler):
    self.conn = conn
    self.handler = handler

    self.conn.sock.settimeout(60.0)
    self.requests = 0
    self.id = httphandler.id
    httphandler.id += 1

    while self.conn.status:
      self.do_request()

  def do_request(self):
    req     = None
    res     = httpresponse()
    error   = None
    finish  = None
    self.requests += 1

    try:
      try:
        req = self.parse_request()
        self.validate_request(req)
        self.process_request(req)

        try:
          finish = self.handler.handle(self, req, res)
        except httperror:
          raise
        except Exception, e:
          error = (500, e)

      except endconnection:
        return

      except httperror, e:
        error = e.args

      if error:
        res = httpresponse()
        res.status = error[0]
        if (res.status <= 100) or (res.status in (204,304)) \
          or (req and req.method == 'HEAD'):
          # these messages cannot have a body
          pass
        else:
          res.body = '%s %s' % (res.status, res.statusmsg)
        try:
          self.handler.handle_error(self, req, res, error[0], *error[1:])
        except: # error handler had an error!
          res = httpresponse()
          res.status = 500
          res.body = '%s %s' % (res.status, res.statusmsg)
          import traceback
          traceback.print_exc()

      self.make_response(req, res, finish)
      self.write_response(res)

      if finish:
        try:   finish()
        except:
          # we already sent out the response+headers,
          # nothing to tell the client at this point
          import traceback
          traceback.print_exc()
          self.conn.close()

      self.finish_response(req, res)

    except IOError, e:
      print e
      self.conn.close()
      return

  def parse_request(self):
    '''
    reads one request from the client and returns it
    '''
    req = httprequest()

    # read the body of the HTTP message, or '' if the client
    # did not specify a C-L
    def read(bytes=None):
      l = cl = int(req.headers.get('Content-Length','0'))
      if not cl: return ''
      data = []
      while l:
        data.append(self.conn.read(l))
        l -= len(data[-1])
      data = ''.join(data)
      assert len(data) == cl, (len(data), cl)
      return data

    req.read = read

    self.readline = self.readrequest
    self._lines = ['']
    try:
      # check if there is a next line to so we can
      # gracefully handle socket closes.
      # we still want to throw an exception if the socket
      # closes when we are reading headers etc
      self._lines.insert(0, self.next_line())
    except EOFError:
      raise endconnection
    while hasattr(self, 'readline'):
      self.readline(req, self.next_line())
    return req

  def next_line(self):
    if len(self._lines[-1]) > 512*1024:
      raise httperror(414) # don't let malicious users use up all the memory

    while len(self._lines) < 2:
      data = self._lines.pop() + self.conn.read()
      self._lines.extend(data.split(CRLF))

    line = self._lines.pop(0)
    return line

  def readrequest(self, request, line):
    if not line: return
    try:
      method, uri, http = line.split(' ')
    except ValueError: raise httperror(400)
    request.method = method
    request.uri = uri
    if   http == 'HTTP/1.0':
      request.http = 1.0
    elif http == 'HTTP/1.1':
      request.http = 1.1
    else:
      raise httperror(505)

    self.readline = self.readheader

  def readheader(self, request, line):
    # TODO: headers can span lines! implement this
    if not line:
      return self.end_headers(request)

    key, val = line.split(':', 1)
    key = key.title()
    val = val.lstrip()

    if key in request.headers:
      #http://www.w3.org/Protocols/rfc2616/rfc2616-sec4.html#sec4.2
      val = request.headers[key] + ', ' + val
    request.headers[key] = val

  def end_headers(self, request):
    # we are done. clean up
    del self.readline
    # if we got extra data, push it back to the front of
    # the connection's read buffer, so someone can read() later
    # this is an abuse of the api because most file objects
    # wont support unread().
    # fix this later.
    last = self._lines.pop()
    l = [i + CRLF for i in self._lines] + [last]
    self.conn.unread(''.join(l))
    del self._lines

  def validate_request(self, req):
    host = req.headers.get('Host')
    if host:
      req.host = host.split(':')[0]
    else:
      req.host = None

    if req.http == 1.1 and not host:
      raise httperror(400)

  def process_request(self, req):
    if req.http == 1.1:
      if req.headers.get('Connection') != 'close':
        # get value from K-A header?
        # set K-A header?
        self.conn.sock.settimeout(300)

  def make_response(self, req, res, finish):
    '''
    create a response object based on the request
    '''
    if not res.http:
      try:     res.http = req.http
      except:  res.http = 1.1

    if not res.status:  res.status = 200
    res.headers.setdefault('Server', 'pyy-httpserver-test')
    res.headers.setdefault('Content-Type', 'text/plain; charset=ISO-8859-4')
    res.headers.setdefault('Date', httptime())

    # http://www.w3.org/Protocols/rfc2616/rfc2616-sec8.html#sec8.1.2
    # default should be to keep the connection open, but this is easier for testing
    res.headers.setdefault('Connection', 'keep-alive')

    if res.body is not None:
      res.body = str(res.body)

    s = []
    if req:
      for k, v in req.headers.iteritems():
        if   k == 'User-Agent': pass
        elif k == 'Host':       pass
        elif k == 'Referer':    pass
        elif k == 'Accept':     pass
        elif k == 'If-Modified-Since':  pass
        elif k == 'Accept-Language':    pass
        elif k == 'Accept-Charset':     pass
        elif k == 'Accept-Encoding':
          res.headers.setdefault('Content-Encoding', 'identity')
          if not res.body:             continue
          if not self.handler.compress:continue
          if not res.compress:         continue
          if finish:                   continue
          tokens = v.lower().split(',')
          len1 = len(res.body)
          if 'deflate' in tokens: # zlib is better, try it first
            if len1 < 32: continue
            import zlib
            res.body = zlib.compress(res.body, 9)
            len2 = len(res.body)
            res.headers['Content-Encoding'] = 'deflate'

          elif 'gzip' in tokens:
            len1 = len(res.body)
            if len1 < 64: continue
            # python's old libraries are retarded.
            # why can't there just be a compress(str) function?
            import gzip, cStringIO
            ss = cStringIO.StringIO()
            gz = gzip.GzipFile(compresslevel=9, mode='wb', fileobj=ss)
            gz.write(res.body)
            gz.flush()
            # end retardedness
            res.body = ss.getvalue()
            len2 = len(res.body)
            res.headers['Content-Encoding'] = 'gzip'

          else: # unsupported encoding. leave it as identity
            continue
            # [C-E] 'identity' ?

          res.headers['Content-Length'] = len2
          s.append('%d/%s (%2.0f%%)' % (len2, len1, 100.*len2/len1))

        elif k == 'Connection':
          if v == 'keep-alive':
            res.headers['Connection'] = v
          else:
            warnings.warn('Unknown Connection token: %s' % v)

        elif k == 'Keep-Alive':
          pass

        else:
          warnings.warn('Unhandled request header: %s: %s' % (k,v))

    if res.body is None:  res.body = ''

    lb = len(res.body)
    if not finish:
      cl = res.headers.get('Content-Length', lb)
      if cl != lb:
        raise Exception('Set C-L header (%d) that was not equal to '
          'response body (%d)' % (cl, lb))
      else:
        res.headers['Content-Length'] = lb

    print '[%d.%d] %d %s %s' % \
        (self.id, self.requests, res.statusnum, req and req.uri, ' '.join(s))
    return res

  def write_response(self, res):
    '''
    write the response object to the client.
    (but maybe not all the content)
    '''
    r = ['HTTP/%.1f %d %s' % (res.http, res.statusnum, res.statusmsg)]
    for k,v in res.headers.iteritems():
      r.append('%s: %s' % (k, v))
    r.append(CRLF)
    self.conn.write(CRLF.join(r))
    self.conn.write(res.body)

  def finish_response(self, req, res):
    '''
    clean up a response. (close the connection if required)
    '''
    con = res.headers.get('Connection')
    if con is None:
      if req.http == 1.0:
        self.conn.close()
    elif con == 'close':
      self.conn.close()
    elif con == 'keep-alive':
      if res.headers.get('Content-Length') is None:
        # TODO don't do this if Transfer-Encoding is chunked
        raise Exception('tried to keep-alive with no C-L')
    else:
      raise Exception('unknown Connecton: token')



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

import re
from urllib import unquote_plus

def parse_query(string):
  d = {}
  string = string.split('&')
  for pair in filter(None, string):
    is_array = False
    try:
      key, value = map(unquote_plus, pair.split('=', 1))
      if key.endswith('[]'): # TODO: why the hell is this a magic string? wtf?
        key   = key[:-2]
        is_array = True
    except ValueError:
      key, value = unquote_plus(pair), ''
    if is_array:
      if key in d:
        d[key].append(value)
      else:
        d[key] = [value]
    else:
      d[key] = value
  return d


def parse_semi(string):
  d = {}
  for pair in filter(None, string.split('; ')):
    try:
      key, value = pair.split('=',1)
    except ValueError:
      key, value = '', pair
    d[key] = value
  return d



def parse_user_agent(user_agent):
  USER_AGENT_REGEX = (
    ('ie',        re.compile(r'MSIE (?P<version>(?P<major>\d+)(?:\.(?P<minor>\d+))?)'), 'Windows CE'),
    ('firefox',   re.compile(r'Firefox/(?P<version>(?P<major>\d+)\.(?P<minor>\d+)(?:\.(?P<release>\d+)(?:\.(?P<build>\d+))?)?((a|b)(?P<beta>\d+))?)'), 'Fennec'),
    ('chrome',    re.compile(r'Chrome/(?P<version>(?P<major>\d+)\.(?P<minor>\d+)(?:\.(?P<release>\d+)(?:\.(?P<build>\d+))?)?)'), None),
    ('safari',    re.compile(r'Version/(?P<version>(?P<major>\d+)\.(?P<minor>\d+)(?:\.(?P<release>\d+))?)'), 'iPhone'),
    ('opera',     re.compile(r'Opera(?:/| )(?P<version>(?P<major>\d+)\.(?P<minor>\d+))'), 'Mini'),
    ('googlebot', re.compile(r'Googlebot/(?P<version>(?P<major>\d+)\.(?P<minor>\d+))'), None), # googlebot Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)
  )
  
  class browser(object): pass
  b = browser()
  
  for browser, regex, mobile in USER_AGENT_REGEX:
    match = re.search(regex, user_agent)
    if match:
      b.name   = browser
      b.mobile = mobile and mobile in user_agent
      
      for name, value in match.groupdict().iteritems():
        if value:
          setattr(b, name, value)
      
      setattr(b, browser, float(b.major + '.' + b.minor))
    else:
      setattr(b, browser, None)
  
  return b



def parse_multipart(content_type, data):
  '''
  This method *MUST* be run with unbuffered input!
  python -u
  '''
  multipart = {}

  content_type = parse_semi(content_type)
  boundary     = '--' + content_type['boundary']
  
  string = data
  if not string.startswith(boundary):
    raise ValueError('form-data boundary does not match')
  
  class part(str): pass
  
  def make_part(string):
    headers     = {}
    disposition = {}
    data        = ''
    
    while True:
      n = string.find('\r\n')
      if n == -1:
        raise ValueError('Unexpected EOF while parsing form parts.')
      if n == 0:
        string = string[2:-2]
        break
      
      header, value = string[:n].split(': ', 1)
      headers[header] = value
      string = string[n+2:]
  
    data        = part(string)
    disposition = parse_semi(headers['Content-Disposition'])
    name        = disposition.pop('name')
    for n, value in disposition.items():
        setattr(data, n, value)
    return name, data
  
  string = data[len(boundary + '\r\n'):]
  
  while True:
    n = string.find(boundary)
    if n == -1:
      raise ValueError('unexpected EOF while parsing form-data (ATTN: make sure this script is running with unbuffered stdin, as the default (text-mode on windows) will fail when reading certain non-text characters) (#python -u for unbuffered stdin/out)')
    name, value = make_part(string[:n])
    multipart[name] = value
    if string[n:-2] == boundary + '--':
        break
    string = string[len(boundary) + n + 2:]

  return multipart
  



'''
Utility classes for creating dynamic html documents
'''

__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import re
from pyy_tag import pyy_tag


def include(f):
  '''
  includes the contents of a file on disk.
  takes a filename
  '''
  fl = file(f, 'rb')
  data = fl.read()
  fl.close()
  return data


def system(cmd, data='', mode='t'):
  '''
  pipes the output of a program
  '''
  import os
  fin, fout = os.popen4(cmd, mode)
  fin.write(data)
  fin.close()
  return fout.read()


def escape(data, quote=True):  # stoled from std lib cgi
  '''
  Escapes special characters into their html entities
  Replace special characters "&", "<" and ">" to HTML-safe sequences.
  If the optional flag quote is true, the quotation mark character (")
  is also translated.

  This is used to escape content that appears in the body of an HTML cocument
  '''
  data = data.replace("&", "&amp;")  # Must be done first!
  data = data.replace("<", "&lt;")
  data = data.replace(">", "&gt;")
  if quote:
    data = data.replace('"', "&quot;")
  return data


_unescape = {
  'quot': 34,
  'amp':  38,
  'lt':   60,
  'gt':   62,
  'nbsp': 32,
  # more here
  # http://www.w3.org/TR/html4/sgml/entities.html
  'yuml': 255,
}


def unescape(data):
  '''
  unescapes html entities. the opposite of escape.
  '''
  cc = re.compile('&(?:(?:#(\d+))|([^;]+));')

  result = []
  m = cc.search(data)
  while m:
    result.append(data[0:m.start()])
    d = m.group(1)
    if d:
      d = int(d)
      result.append(d > 255 and unichr(d) or chr(d))
    else:
      d = _unescape.get(m.group(2), ord('?'))
      result.append(d > 255 and unichr(d) or chr(d))

    data = data[m.end():]
    m = cc.search(data)

  result.append(data)
  return ''.join(result)


_reserved = ";/?:@&=+$, "
_replace_map = dict((c, '%%%2X' % ord(c)) for c in _reserved)


def url_escape(data):
  return ''.join(_replace_map.get(c, c) for c in data)


def url_unescape(data):
  return re.sub('%([0-9a-fA-F]{2})',
    lambda m: chr(int(m.group(1), 16)), data)


class lazy(pyy_tag):
  '''
  delays function execution until rendered
  '''
  def __init__(self, func, *args, **kwargs):
    pyy_tag.__init__(self)

    self.func   = func
    self.args   = args
    self.kwargs = kwargs

  def render(self, indent=1, inline=False):
    return self.func(*self.args, **self.kwargs)


# TODO rename this to raw?
class text(pyy_tag):
  '''
  Just a string. useful for inside context managers
  Note: this will not escape HTML, it is a raw passthrough
  '''
  def __init__(self, text):
    pyy_tag.__init__(self)
    self.text = text

  def render(self, indent, inline):
    return self.text



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import os
from pyy.server import fileserver

class syntaxfileserver(fileserver):
  def write_file(self, conn, req, res, path, *args):
    try:
      from pygments            import highlight
      from pygments.lexers     import get_lexer_for_filename
      from pygments.formatters import HtmlFormatter

    except ImportError:
      import warnings
      warnings.warn('pygments not found, serving files as plain-text')
      return fileserver.write_file(self, conn, req, res, path, *args)

    else:
      code = file(path).read()
      name = os.path.basename(path)

      ret = highlight(code, get_lexer_for_filename(name), HtmlFormatter(full=True))
      res.headers['Content-Type']   = 'text/html'
      res.body = ret





__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import stackless
import threading

class threadio(object):
  def __init__(self, file):
    self.file = file

  def __getattr__(self, attr):
    f = getattr(self.file, attr)
    if not callable(f):
      return f
    def wrapper(*args, **kwargs):
      ch = stackless.channel()
      def thread():
        try:
          r = f(*args, **kwargs)
          ch.send(r)
        except Exception, e:
          ch.send_exception(type(e), e)
      threading.Thread(target=thread).start()
      return ch.receive()
    return wrapper





__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import stackless
import socket
import select
import threading
import time

class connection(object):
  READABLE = 1
  WRITABLE = 2
  OPEN = 1

  def __init__(self, server, sock, addr):
    self.server   = server
    self.addr     = addr
    self.sock     = sock
    self.handler  = None

    self.readbuffer  = []
    self.writebuffer = []

    self.readchannel  = stackless.channel()
    self.writechannel = stackless.channel()

    self.status = connection.OPEN

  def fileno(self):
    return self.sock.fileno()

  def onreadable(self):
    if not self.status: return
    try:
      data = self.sock.recv(256*1024)
    except socket.error, e:
      if  e.args[0] == 10053  or \
          e.args[0] == 10054  or \
          e.args[0] == 9      or \
          e.args[0] == 105    or \
          e.args[0] == 104:         # Connection reset by peer
        #print 'Client %s closed the connection' % (self.addr,)
        self.onclose(e)
        return
      if e.args[0] == 10035:  # would have blocked
        return
      if e.args[0] == 11:     # resource temp unavailable
        return
      else: raise

    if data == '':
      # TODO
      # This happens when the client sent a FIN.
      # It might still be listening for data, however!
      #print 'Client %s closed the connection' % (self.addr,)
      self.onclose()
      return

    # print '%s GOT DATA: %r' % (self.addr, data)
    self.readbuffer.append(data)
    if self.handler and self.handler.hasattr('ondata'):
      self.handler.ondata(data)

    # is someone waiting on read?
    if self.readchannel.balance < 0:
      self.readchannel.send(None)

  def onwritable(self):
    if not self.status: return
    while self.writebuffer:
      d = self.writebuffer[0]
      o = d
      #print '%s SENDING: %r' % (self.addr, d[:80])

      try:
        while d:
          d = d[self.sock.send(d):]
          self.writebuffer[0] = d
        self.writebuffer.pop(0)
        self.writechannel.send(o)

      except socket.error, e:
        if e.args[0] == 10053:  # closed the connection
          self.onclose(e)
          return
        if e.args[0] == 10035:  # would have blocked
          return
        if e.args[0] == 11:     # resource temp unavailable
          return
        else: raise

    self.server.set_writable(self, False)

  def onclose(self, e=None):
    if not self.status: return # we closed already
    while self.readchannel.balance < 0:
      self.readchannel.send_exception(EOFError, EOFError('connection reset', e))
    while self.writechannel.balance < 0:
      self.writechannel.send_exception(EOFError, EOFError(e))

    self.server.close(self)
    self.status = 0
    if self.handler and self.handler.hasattr('onclose'):
      self.handler.onclose()

  def write(self, data):
    if not self.status:
      raise EOFError('write on closed connection')
    assert type(data) == str
    #print 'WRITE: %r' % data[:80]
    self.writebuffer.append(data)
    self.server.set_writable(self)
    # block until it sent
    r = self.writechannel.receive()
    #assert data == r, (len(data), len(r), data[:80], r[:80])
    assert data.endswith(r) # this is not a very strong check, but
                            # the one above fails when the loop on
                            # line 87 is interrupted and o != data
                            # on the next onwritable() call
    return r

  def read(self, x=None):
    if not self.status:
      raise EOFError('read on closed connection')
    if not any(self.readbuffer):
      self.server.set_readable(self)
      # block until data availible
      self.readchannel.receive()
    data = ''.join(self.readbuffer)
    self.readbuffer = x and [data[x:]] or []
    return data[:x]

  def close(self, how=socket.SHUT_RDWR):
    #print 'CLOSING: %s' % (self.addr,)
    if self.writebuffer:
      #self.write('')
      raise Exception('attempt to close connection with unsent data', self.writebuffer)
    self.sock.shutdown(how)
    self.sock.close()
    self.onclose()


class listener(connection):
  def __init__(self, server, sock, addr, handler, *args):
    connection.__init__(self, server, sock, addr)
    self.handler  = handler
    self.args     = args
    self.server.set_readable(self)

  def onreadable(self):
    while 1:
      try:
        sock, addr = self.sock.accept()
      except socket.error, e:
        if e.args[0] == 10035: # would have blocked
          return
        if e.args[0] == 11:     # resource temp unavailable
          return
        else: raise

      sock.setblocking(0)
      conn = connection(self.server, sock, addr)
      self.server.requests.add(conn)
      handler = stackless.tasklet(self.handler)(self.server, conn, *self.args)
      #print 'Accepted connection: %s' % (addr,)


class handler(object):
  def __init__(self, server, conn, *args):
    self.conn = conn
    self.server = server


class server(object):
  def __init__(self):
    self.listeners = []
    self.requests  = set()
    self.readable  = set()
    self.writable  = set()
    self.selectch  = stackless.channel()
    self.done      = False

  def listen(self, addr, handler, *args):
    sock = socket.socket()
    sock.setblocking(0)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(addr)
    sock.listen(100)
    self.listeners.append(listener(
      self, sock, addr, handler, *args))
    print 'Listening on %s' % (addr,)

  def set_readable(self, req, x=True):
    if x: self.readable.add(req)
    else: self.readable.discard(req)
    if x: self.interrupt()

  def set_writable(self, req, x=True):
    if x: self.writable.add(req)
    else: self.writable.discard(req)
    if x: self.interrupt()

  def close(self, req):
    self.readable.discard(req)
    self.writable.discard(req)
    self.requests.discard(req)

  def interrupt(self):
    if self.selectch.balance < 0:
      self.selectch.send(([],[],[]))

  def select(self, timeout):
    if timeout==0:
      r, w, e = select.select(
        self.readable, self.writable, [], timeout)

      return r, w, e

    def thread():
      r = select.select(
        self.readable, self.writable, [], timeout)
      #if self.selectch.balance < 0: # this causes stackless to crash!
      self.selectch.send(r)

    threading.Thread(target=thread).start()

    r = self.selectch.receive()
    while self.selectch.balance > 0:
      r2 = self.selectch.receive()
      map(lambda x:reduce(list.extend,x) ,zip(r,r2))
    return r

  def tick(self, timeout=None):
    readable, writable, _e = self.select(timeout)

    for i in readable:
      i.onreadable()

    for i in writable:
      i.onwritable()

  def run(self, timeout=None):
    def mainloop():
      try:
        while 1:
          self.tick(timeout)
          stackless.schedule()
          time.sleep(0.05)
      except KeyboardInterrupt: pass
      finally:
        self.quit(timeout)

    stackless.tasklet(mainloop)()
    stackless.run(threadblock=True)

  def quit(self, timeout):
    self.done = True
    print 'Quiting'
    time.sleep(timeout or 0)
    while self.selectch.balance > 0:
      self.selectch.receive()

def main():
  s = server()
  s.listen(('',1234), handlers.echohandler)
  s.run()

if __name__ == '__main__':
  main()






from pyy.html import document
from pyy.html.tags import *

class page(document):
  prefix  = ''
  css     = []
  script  = []
  inline  = False

  def __init__(self, **kwargs):
    document.__init__(self, **kwargs)

    for i in self.css:
      if self.inline:
        self.head += style(include(i), type="text/css")
      else:
        if not i.startswith('http://'):
          i = self.prefix + i
        self.head += link(href=i, type="text/css", rel="stylesheet")

    for i in self.script:
      if self.inline:
        self.head += script(include(i), type="text/javascript")
      else:
        if not i.startswith('http://'):
          i = self.prefix + i
        self.head += script(src=i, type="text/javascript")


  @classmethod
  def get(cls, handler ,req, res, *args, **kwargs):
    p = cls()
    p.do_get(*args, **kwargs)
    res.headers['Content-Type'] = 'text/html'
    res.body = p.render()

  @classmethod
  def post(cls, handler ,req, res, *args, **kwargs):
     raise httperror(405)

  @classmethod
  def handle(cls, handler, req, res, *args, **kwargs):
    h = getattr(cls, req.method.lower(), None)
    if not h: raise httperror(405)
    h(handler, req, res, *args, **kwargs)

  @classmethod
  def handle_error(cls, handler, req, res, status, *args):
    p = cls()
    h = getattr(p, 'do_%d' % status, getattr(p, 'do_error', None))
    if not h: raise httperror(405)
    h(status, *args)
    res.headers['Content-Type'] = 'text/html'
    res.body = p.render()



#!/usr/bin/env python

__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

#Ensure we get the repository version and not an installed version
import os
import sys
sys.path.insert(0, os.path.split(os.path.dirname(os.path.abspath(__file__)))[0])

from unittest import defaultTestLoader, TestSuite, TextTestRunner
import tests.html, tests.server, tests.web

#Valid modules to test
VALID = {'html'  : 1,
         'server': 2,
         'web'   : 4,
         'all'   : 7}

#Default to all modules unless any arguments were passed in
pkgs = 0 if len(sys.argv[1:]) else VALID['all']

#Load modules from arguments
for arg in sys.argv[1:]:
  if arg not in VALID:
    print 'INVALID MODULE: %s.' % arg
  else:
    tests |= VALID[arg]

#Assemble individual module tests into array
suites = []
if pkgs & VALID['html']:   suites += defaultTestLoader.loadTestsFromModule(tests.html)
if pkgs & VALID['server']: suites += defaultTestLoader.loadTestsFromModule(tests.server)
if pkgs & VALID['web']:    suites += defaultTestLoader.loadTestsFromModule(tests.web)

#Run tests on each module in the array

def test(verbosity=2):
  return TextTestRunner(verbosity=verbosity).run(TestSuite(suites))

if __name__ == '__main__':
	test()




__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

import unittest
import os
import test
from pyy.web         import httperror
from pyy.web.resolve import RegexResolver, FileHeirarchyResolver

path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'test')


class Request:
  def __init__(self, uri):
    self.uri = uri


class ResolverTests(unittest.TestCase):
  def testSelectIndexInFolder(self):
    p = FileHeirarchyResolver(path, Request('admin/'))
    self.assertEqual(str(p.__class__), "<class '__init__.Index'>")
  
  def testSelectIndexInFile(self):
    p = FileHeirarchyResolver(path, Request('admin/a/'))
    self.assertEqual(str(p.__class__), "<class 'a.Index'>")
  
  def testSelectClassInFile(self):
    p = FileHeirarchyResolver(path, Request('admin/a/test'))
    self.assertEqual(str(p.__class__), "<class 'a.Test'>")
  
  def testErrorWhenNoIndex(self):
    try:
      FileHeirarchyResolver(path, Request('admin/hi/'))
      self.fail()
    except httperror:
      pass



__license__ = '''
This file is part of pyy.
 
pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.
 
pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.
 
You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

import unittest
from pyy.html.document import document
from pyy.html.tags     import body, h1, p, comment
from pyy.html.util     import *


class RenderingTests(unittest.TestCase):
  def testInline(self):
    self.assertEqual(str(p(h1(), __inline=True)), '<p><h1></h1></p>')
  
  def testIndented(self):
    self.assertEqual(str(p(h1())), '<p>\n\t<h1></h1>\n</p>')
  
  def testIndentedChildren(self):
    self.assertEqual(str(body(p(), p())), '<body>\n\t<p></p>\n\t<p></p>\n</body>')
  
  def testComment(self):
    self.assertEqual(str(comment('test')), '<!--test-->')
  
  def testCommentWithTags(self):
    self.assertEqual(str(body(p(), comment(p()))), '<body>\n\t<p></p>\n\t<!--\n\t<p></p>\n\t-->\n</body>')
  
  def testConditionalComment(self):
    self.assertEqual(str(comment(p(), condition='lt IE 7')), '<!--[if lt IE 7]>\n<p></p>\n<![endif]-->')
  
  def testIndentedConditionalComment(self):
    self.assertEqual(str(body(p(), comment(p(), condition='lt IE 7'))), '<body>\n\t<p></p>\n\t<!--[if lt IE 7]>\n\t<p></p>\n\t<![endif]-->\n</body>')
  
  def testDocumentTitleUpdate(self):
    d1 = document()
    rd1 = d1.render()
    d1.title = "test"
    rd1 = d1.render()
    d2 = document(title="test")
    rd2 = d2.render()
    self.assertEqual(rd1, rd2)

  def testEscape(self):
    self.assertEqual(str(p('Hi & There')), '<p>Hi & There</p>')
    self.assertEqual(str(p(escape('Hi & There'))), '<p>Hi &amp; There</p>')
    self.assertEqual(str(p(unescape('Hi &amp; There'))), '<p>Hi & There</p>')
    
    



'''
About
=====
`pyy.httpserver` is a lightweight, highly configurable web server written using
stackless python for ultimate effeciency.

__`pyy.httpserver` is still considered highly experimental.__


Usage
=====
`pyy.httpserver` allows you to serve files with different pluggable modules
based upon a variety of conditions.

The server is configured with a series of specifiers and handlers arranged in
a list:
    [ SPECIFIER, HANDLER, ... ]

`SPECIFIER` must be one of:

* A regular expression string will match that hostname (or uri, in a sub-handler).
* An HTTP error code matches that code.
* `httperror` matches all errors.
* `None` will match `None` or the empty string.
* `True` will match everything.

If no specifiers match, an httperror(404) will be returned.

If multiple specifiers match, the last one will be used. If the last one
returns a 404 error, then the second last one will be used, etc.

*Note that if a leaf-node (final handler) raises a 404, that error will be
handled normally and will not cause the resolver to keep looking.*

`HANDLER` must be one of:

* `[ SPECIFIER, HANDLER, ...]` to do further processing.
* A function object will be called with `f(*args)` where `args` is a list of all regex matches that were encountered in the entire processing of this request. The function's return value will then be used as a new `HANDLER`.
* A handler object will be called with `.handle(httpserver, request, response, *args)` (or `.handle_error(..)` if this is an error match).
* `None` to do nothing.
* An `httperror` instance to raise that error.

You can see an example server configuration [here](http://github.com/Knio/pyy/blob/master/docs/example/run.py).


Developed By
============
* Tom Flanagan - <theknio@gmail.com>
* Jake Wharton - <jakewharton@gmail.com>

Git repository located at
[github.com/Knio/pyy](http://github.com/Knio/pyy)


License
=======
    Copyright 2009 Tom Flanagan, Jake Wharton

    This file is part of pyy.

    pyy is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    pyy is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with pyy.  If not, see
    <http://www.gnu.org/licenses/>.
'''

__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

# from server             import server
from httpserver         import httpserver
from fileserver         import *
from filterserver       import *
from pyyserver          import *
from syntaxfileserver   import *



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

'''

This module implements a threaded socket server

'''

import socket
import threading
import time

class connection(object):
  '''
  A connection object. Wraps socket

  Public API:

  .read(n)      returns n bytes
  .write(data)  writes data
  .close([how]) closes the socket

  .unread(data) pushes data back onto read buffer
  '''


  OPEN = 1
  CLOSED = 0

  def __init__(self, server, sock, addr):
    self.server   = server
    self.addr     = addr
    self.sock     = sock

    self.readbuffer  = []

    self.status = connection.OPEN

  def fileno(self):
    return self.sock.fileno()

  def read(self, n=None):
    if not n:
      if self.readbuffer:
        data = ''.join(self.readbuffer)
        self.readbuffer = []
        return data

      else:
        while 1:
          try:
            data = self.sock.recv(1024*1024)
            if data == '':
              self.status = connection.CLOSED
              raise EOFError()
            return data
          except socket.error, e:
            # resource temp unavailable
            if e.args[0] == 11: time.sleep(0.1)
            else: raise


    while sum(map(len, self.readbuffer)) < n:
      try:
        data = self.sock.recv(1024*1024)
        if data == '':  raise EOFError()
        self.readbuffer.append(data)
      except socket.error, e:
        # resource temp unavailable
        if e.args[0] == 11: time.sleep(0.1)
        else: raise

    data = ''.join(self.readbuffer)

    self.readbuffer = [data[n:]]
    return data[:n]

  def write(self, data):
    while data:
      try:
        data = data[self.sock.send(data):]
      except socket.error, e:
        # resource temp unavailable
        if e.args[0] == 11: time.sleep(0.1)
        else: raise


  def close(self, how=socket.SHUT_RDWR):
    self.sock.shutdown(how)
    self.sock.close()
    self.status = 0

  def unread(self, data):
    self.readbuffer[0:0] = [data]


class server(object):
  '''
  A socket server

  .listen(addr, handler, *a, **kw)

  Listen on addr. When a connetion is accepted,
  call handler(server, connection, *a, **kw)
  '''


  def __init__(self):
    self.listenthreads = []
    self.done = False

  def listen(self, addr, handler, *args, **kwargs):
    sock = socket.socket()
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.settimeout(1.0)
    sock.bind(addr)
    sock.listen(5)

    def listen():
      while not self.done:
        try:
          newsock, newaddr = sock.accept()
        except socket.timeout:
          time.sleep(0.1)

        else:
          newsock.settimeout(5.0)
          conn = connection(self, newsock, newaddr)

          def handle():
            try:
              handler(self, conn, *args, **kwargs)
            finally:
              newsock.close()

          def profile():
            h = handle
            import cProfile as profile
            profile.runctx("h()", globals(), locals())

          thread = threading.Thread(target=handle)
          thread.daemon = True
          thread.start()

    listenthread = threading.Thread(target=listen)
    listenthread.start()
    self.listenthreads.append(listenthread)
    print('Listening on %r' % (addr,))

  def quit(self):
    if self.done: return
    print('Quitting server')
    self.done = True

  def run(self):
    try:
      while not self.done:
        time.sleep(5.0)

    finally:
      self.quit()

def main():
  '''
  Sample echo server
  '''

  def echo(serv, conn):
    while 1:
      conn.write(conn.read())

  s = server()
  s.listen(('',1234), echo)
  s.run()

if __name__ == '__main__':
  main()







__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

import unittest
from pyy.html.tags import *
from pyy.html.util import *


class AttributeTests(unittest.TestCase):
  def testAddViaDict(self):
    i = img()
    i['src'] = 'test.png'
    self.assertEqual(str(i), '<img src="test.png">')

  def testAddViaKeywordArg(self):
    i = img(src='test.png')
    self.assertEqual(str(i), '<img src="test.png">')

  def testBooleanAttribute(self):
    i = img(test=True)
    self.assertEqual(str(i), '<img test="test">')

  def testUtils(self):
    d = div()
    d += system('echo hi')
    self.assertEqual(str(d), '<div>hi\n</div>')




__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

from tags import html, body, head, title

class document(object):
  def __init__(self, title='PYY Page', doctype='<!DOCTYPE html>', request=None):
    '''
    Creates a new document instance. Accepts `title`, `doctype`, and `request` keyword arguments.
    '''
    self.cookies       = {}
    self.doctype       = doctype
    self.html          = html()
    self.html.document = self
    self.head          = self.html.add(head())
    self.body          = self.html.add(body())
    self._entry        = self.body
    self.title         = title
    self.request       = request

  def add(self, obj):
    '''
    Adding tags to a document appends them to the <body>.
    '''
    return self._entry.add(obj)

  def __iadd__(self, obj):
    self._entry += obj
    return self


  def _get_title(self):
    if title not in self.head:
      self.head += title('PYY Page')
    return self.head.get(title)[0].children[0]

  def _set_title(self, value):
    if title in self.head:
      self.head.get(title)[0].children = [value]
    else:
      self.head += title(value)

  title = property(_get_title, _set_title, None, 'Document title.')

  def __enter__(self):
    self._entry.__enter__()
    return self

  def __exit__(self, *args):
    return self._entry.__exit__(*args)

  def validate(self):
    '''
    Validates the tag tree structure and its attributes.
    '''
    self.html.validate()

  def render(self, *args, **kwargs):
    '''
    Creates a <title> tag if not present and renders the DOCTYPE and tag tree.
    '''
    r = []

    #Validates the tag tree and adds the doctype if one was set
    if self.doctype:
      r.append(self.doctype)
      r.append('\n')
    r.append(self.html.render(*args, **kwargs))

    return u''.join(r)
  __str__ = __unicode__ = render

  def __repr__(self):
    return '<pyy.html.document "%s">' % self.title



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import parsers

class httperror(Exception): pass

class httpmessage(object):
  def __init__(self):
    self.http    = 0
    self.headers = {}
    self.body    = None

class httprequest(httpmessage):
  def __init__(self):
    httpmessage.__init__(self)
    self.method = None
    self.uri    = None
    del self.body

  def read(self):
    raise NotImplementedError

  def _get(self):
    try:    get = self.uri.split('?', 1)[1]
    except: get = ''
    return parsers.parse_query(get)

  def _cookie(self):
    return parsers.parse_semi(self.headers.get('Cookie', ''))

  def _post(self):
    ct = self.headers.get('Content-Type','')
    if ct.startswith('multipart/form-data'):
      return parsers.parse_multipart(ct, self.body)
      
    elif ct.startswith('application/x-www-form-urlencoded'):
      return parsers.parse_query(self.body)

    else:
      return {}

  def _ua(self):
    return parsers.parse_user_agent(self.headers.get('User-Agent', ''))

  def _body(self):
    return self.read()

  def __getattr__(self, attr):
    if not hasattr(self, '_' + attr):
      raise AttributeError

    val = getattr(self, '_' + attr)()
    setattr(self, attr, val)
    return val

class httpresponse(httpmessage):
  def __init__(self):
    httpmessage.__init__(self)
    self.statusnum = None
    self.statusmsg = None
    self.compress  = True
    self.cookies   = {}

  def set_cookie(self, cookie):
    self.cookies[cookie.name] = cookie
    self.headers['Cookie'] = ' '.join(c.render() for c in self.cookies.values())
  
  def write(self, data):
    raise NotImplementedError

  def set_status(self, x):
    self.statusnum = x
    self.statusmsg = httpresponse.STATUS.get(x, 'OK')
  
  status = property(lambda s: s.statusnum, set_status)
  
  STATUS = {
    100: "Continue",
    101: "Switching Protocols",
    200: "OK",
    201: "Created",
    202: "Accepted",
    203: "Non-Authoritative Information",
    204: "No Content",
    205: "Reset Content",
    206: "Partial Content",
    300: "Multiple Choices",
    301: "Moved Permanently",
    302: "Found",
    303: "See Other",
    304: "Not Modified",
    305: "Use Proxy",
    307: "Temporary Redirect",
    400: "Bad Request",
    401: "Unauthorized",
    402: "Payment Required",
    403: "Forbidden",
    404: "Not Found",
    405: "Method Not Allowed",
    406: "Not Acceptable",
    407: "Proxy Authentication Required",
    408: "Request Time-out",
    409: "Conflict",
    410: "Gone",
    411: "Length Required",
    412: "Precondition Failed",
    413: "Request Entity Too Large",
    414: "Request-URI Too Large",
    415: "Unsupported Media Type",
    416: "Requested range not satisfiable",
    417: "Expectation Failed",
    500: "Internal Server Error",
    501: "Not Implemented",
    502: "Bad Gateway",
    503: "Service Unavailable",
    504: "Gateway Time-out",
    505: "HTTP Version not supported",
  }



'''
About
=====
`pyy.html` allows for creating (X)HTML markup through the use of objects.
This allows you to tightly integrate (X)HTML generation into your backend
without the need of using an intermediate templating language.

`pyy.html` also provides you with helper classes for generating and parsing
(X)HTML documents.


Usage
=====
All these examples assume you have imported the appropriate tags or entire tag
set (i.e. `from pyy.html.html import *`).

Hello, pyy!
-----------
Constructing a "Hello, World!"-style example is as easy as this:
    >>> print html(body(h1('Hello, pyy!')))
    <html>
      <body>
        <h1>Hello, pyy!</h1>
      </body>
    </html>


Complex Structures
------------------
Through the use of the `+=` operator and the `.add()` method you can easily
create more advanced structures.

Create a simple list:
    >>> list = ul()
    >>> for item in range(4):
    >>>   list += li('Item #', item)
    >>> print list
    <ul>
      <li>Item #0</li>
      <li>Item #1</li>
      <li>Item #2</li>
      <li>Item #3</li>
    </ul>

If you are using a database or other backend to fetch data, `pyy.html` supports
iterables to help streamline your code:
    >>> print ul(li(a(name, href=link), __inline=True) for name, link in menu_items)
    <ul>
      <li><a href="/home/">Home</a></li>
      <li><a href="/about/">About</a></li>
      <li><a href="/downloads/">Downloads</a></li>
      <li><a href="/links/">Links</a></li>
    </ul>

A simple document tree:
    >>> _html = html()
    >>> _body = _html.add(body())
    >>> header  = _body.add(div(id='header'))
    >>> content = _body.add(div(id='content'))
    >>> footer  = _body.add(div(id='footer'))
    >>> print _html
    <html>
      <body>
        <div id="header"></div>
        <div id="content"></div>
        <div id="footer"></div>
      </body>
    </html>

For clean code, the `.add()` method returns children in tuples. The above
example can be cleaned up and expanded like this:
    >>> _html = html()
    >>> _head, _body = _html.add(head(title('Simple Document Tree')), body())
    >>> names = ['header', 'content', 'footer']
    >>> header, content, footer = _body.add(div(id=name) for name in names)
    >>> print _html
    <html>
      <head>
        <title>Simple Document Tree</title>
      </head>
      <body>
        <div id="header"></div>
        <div id="content"></div>
        <div id="footer"></div>
      </body>
    </html>

You can modify the attributes of tags through a dictionary-like interface:
    >>> header = div()
    >>> header['id'] = 'header'
    >>> print header
    <div id="header"></div>

Comments can be created using objects too!
    >>> print comment('BEGIN HEADER')
    <!--BEGIN HEADER-->
    >>> print comment(p('Stop using IE5!'), condition='lt IE6')
    <!--[if lt IE6]>
    <p>Stop using IE5!</p>
    <![endif]-->

Creating Documents
------------------
Since creating the common structure of an HTML document everytime would be
excessively tedious pyy.html provides a class to create an manage them for
you, `document`.

When you create a new document, the basic HTML tag structure is created for
you.
    >>> d = document()
    >>> print d
    <html>
      <head>
        <title>PYY Page</title>
      </head>
      <body></body>
    </html>

The `document` class also provides helpers to allow you to access the `html`,
`head`, and `body` elements directly.
    >>> d = document()
    >>> d.html
    <pyy.html.html.html: 0 attributes, 2 children>
    >>> d.head
    <pyy.html.html.head: 0 attributes, 0 children>
    >>> d.body
    <pyy.html.html.body: 0 attributes, 0 children>

You should notice that here the `head` tag contains zero children. This is
because the default `title` tag is only added when the document is rendered
and the `head` element already does not explicitly contain one.

The `document` class also provides helpers to allow you to directly add
elements to the `body` tag.
    >>> d = document()
    >>> d += h1('Hello, World!')
    >>> p += p('This is a paragraph.')
    >>> print d
    <html>
      <head>
        <title>PYY Page</title>
      </head>
      <body>
        <h1>Hello, World!</h1>
        <p>This is a paragraph.</p>
      </body>
    </html>

Markup Validation
-----------------
You can also set the DOCTYPE of the `document` which will validate the tag tree
when it is rendered.
    >>> d = document()
    >>> d.doctype = dtd.xhtml11
    >>> print d
    <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN" "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
    <html xmlns="http://www.w3.org/1999/xhtml">
      <head>
        <title>PYY Page</title>
      </head>
      <body></body>
    </html>

Notice how the required XHTML 1.1 attribute `xmlns` is automatically added to
the `<html>` tag.

If there are any errors in the tag tree a `ValueError` will be thrown which
describes the offending tag. You can check the validity of your document at
any time by calling `validate()` explicity.

You may also set a `document`'s DOCTYPE with the keyword argument `doctype`.
    >>> d = document(doctype=dtd.html5)
However, doing so will cause the entire document to be validated with every
added tag. For production environments it is best to set the DOCTYPE after the
entire tag tree has been created so it all is only validated once.

Parsing Documents
-----------------
The `parser` class contains two methods: `parse` and `pageparse`. `parse` will
take valid and mostly-valid (X)HTML input and return a tag tree. `pageparse`
will take an entire document and return a `document` instance complete with
the DOCTYPE (if present) and a tag tree using the (X)HTML version specified
by the DOCTYPE.

`parse` will simply return a heirarchy of all the tags and their content that
it can recognize in a string.
    >>> parse('<p>Hello.</p>')
    <pyy.html.html.p: 0 attributes, 1 child>
    >>> parse('<html><head><title>test</title></head><body><h1>Hi.</h1></body></html>')
    <pyy.html.html.html: 0 attributes, 2 children>
    >>> parse('<div id="first"></div><div id="second"></div>')
    [<pyy.html.html.div: 1 attribute, 0 children>, <pyy.html.html.div: 1 attribute, 0 children>]

Notice that if multiple top-level tags exist in the string `parse` will return
them as an array.

`pageparse` also takes a string of tags and optionally a DOCTYPE and returns a
`document` object.
    >>> pageparse('<!DOCTYPE html><html><head><title>Test</title></head><body></body></html>')
    <pyy.html.document.document html5 "Test">


Developed By
============
* Tom Flanagan - <theknio@gmail.com>
* Jake Wharton - <jakewharton@gmail.com>

Git repository located at
[github.com/Knio/pyy](http://github.com/Knio/pyy)


License
=======
    Copyright 2009 Tom Flanagan, Jake Wharton

    This file is part of pyy.

    pyy is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    pyy is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with pyy.  If not, see
    <http://www.gnu.org/licenses/>.
'''

from document import document



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import datetime

class cookie(object):
  def __init__(self, name, value, expires=None, duration=None, path='/',
               domain=None, secure=False, httponly=False):
    
    if duration and expires:
        raise ValueError('Cookie cannot have both expiry date and duration.')
    
    self.name     = name.lower()
    self.value    = value
    self.path     = path
    self.expires  = expires
    self.duration = duration
    self.domain   = domain
    self.secure   = secure
    self.httponly = httponly

  def render(self, is_header=False):
    timestamp = None
    if self.expires:
      if isinstance(self.expires, datetime.datetime):
        timestamp = self.expires
      elif isinstance(self.expires, float) or \
           isinstance(self.expires, int):
        timestamp = datetime.datetime.utcfromtimestamp(self.expires)
      else:
        raise TypeError('Invalid expiry time.')
    if self.duration:
      timestamp = datetime.datetime.utcnow()
      if isinstance(self.duration, float) or \
          isinstance(self.duration, int):
         timestamp += datetime.timedelta(seconds=self.duration)
      else:
        raise TypeError('Invalid duration.')
     
    cookie = []
    cookie.append('%s=%s;' % (self.name, self.value))
    cookie.append('path=%s;' % self.path)
    if timestamp:     cookie.append('expires=%s;' % timestamp.strftime("%a, %d-%b-%Y %H:%M:%S GMT"))
    if self.domain:   cookie.append('domain=%s;' % self.domain)
    if self.secure:   cookie.append('secure;')
    if self.httponly: cookie.append('httponly;')
    
    return ('Set-Cookie: ' if is_header else '') + ' '.join(cookie)
  
  def __str__(self):
    return self.render()
  __unicode__ = __str__
  
  def __repr__(self):
    return "<pyy.web.cookie.cookie %s>" % self.render()



# wsgi.py - pyy WSGI gateway
#
# WARNING: Edit this file very carefully, a bug will crash your site!

import routing

def application(environ, start_response):
  method = environ['REQUEST_METHOD']
  uri = environ['PATH_INFO']

  try:
    #TODO check return value, pass args
    body = str(routing.match(uri, method)())
    status = '200 OK'

  except SystemExit:
    pass

  except Exception, e:
    import traceback
    body = '''<!DOCTYPE html>
<html><head><title>Error</title></head><body><pre>
An error occured while generating this page:

%s
</pre></body></html>''' % traceback.format_exc()
    status = '500 INTERNAL SERVER ERROR'

  headers = [
    ('Content-Type', 'text/plain'),
    ('Content-Length', str(len(body))),
  ]
  start_response(status, headers)

  return body



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

# import stackless
import os
import os.path
import sys
import http
import time
import warnings
from datetime import datetime
from urllib   import unquote_plus
from pyy.web  import httperror


class staticserver(object):
  def __init__(self, handler):
    self.handler  = handler

  def handle(self, conn, req, res, *args):
    if not req.method in ['GET', 'HEAD']:
      raise httperror(405)

    cache     = True
    modified  = True
    for k,v in req.headers.iteritems():
      if k == 'If-Modified-Since':
        try:     t = http.httptime(v)
        except:  continue # malformed header value
        if self.handler.get_mtime(*args) <= t: # I hate time libraries so much
          modified = False

      elif k == 'Cache-Control':
        if v == 'no-cache':
          cache = False

      elif k == 'Expect':   raise httperror(417)
      elif k == 'If-Range': raise httperror(501)
      elif k == 'Range':    raise httperror(501)

      elif k == 'Referrer': pass

      else:
        warnings.warn('unhandled request header: %s: %s' % (k,v))

    if not modified and cache:
      raise httperror(304)

    return self.handler.handle(conn, req, res, *args)


# from stacklessfileIOCP import stacklessfile

class fileserver(object):
  MIME = {
  'txt':  'text/plain',
  'htm':  'text/html',
  'html': 'text/html',
  'pyc':  'application/octet-stream',
  'css':  'text/css',
  'jpg':  'image/jpeg',
  'jpeg': 'image/jpeg',
  'png':  'image/png',
  }

  COMPRESS = {
  'png':  False,
  'jpg':  False,
  'jpeg': False,
  'gif':  False,
  'zip':  False,
  'rar':  False,
  'mp3':  False,
  'avi':  False,
  }

  def __init__(self, root='.', dir_listing=True):
    self.root = root
    self.dir_listing = dir_listing

  def check_path(self, path):
    if path.startswith('/'):
      path = path[1:]

    path = os.path.join(self.root, unquote_plus(path))
    path = os.path.realpath(os.path.abspath(path))
    root = os.path.realpath(os.path.abspath(self.root))

    # don't let them go outside root
    if not path.startswith(root):
      raise httperror(403)

    # does it exist?
    if not os.path.exists(path):
      raise httperror(404)

    valid = False
    if os.path.isfile(path): valid = True
    if os.path.isdir(path) and self.dir_listing: valid = True

    if not valid:
      raise httperror(406)

    return path

  def get_mtime(self, *args):
    path = self.check_path(*args)
    return datetime.utcfromtimestamp(os.path.getmtime(path))

  def handle(self, conn, req, res, path, *unused):
    path = self.check_path(path)

    if os.path.isfile(path):
      fsize = os.path.getsize(path)
      mtime = os.path.getmtime(path)
      try:    ext = os.path.basename(path).split('.')[-1]
      except: ext = ''
      res.compress  = self.COMPRESS.get(ext, True)
      res.headers['Content-Type']   = self.MIME.get(ext, 'text/plain')
      res.headers['Content-Length'] = fsize
      res.headers['Last-Modified']  = http.httptime(time.gmtime(mtime))

      if req.method == 'GET':
        return self.write_file(conn, req, res, path)

    if os.path.isdir(path):
      # what if HEAD?
      res.headers['Content-Type'] = 'text/html'
      res.body = self.write_dir_listing(path)


  def write_file(self, conn, req, res, path):
    fsize = os.path.getsize(path)

    f = file(path, 'rb')

    # res.body = f.read()
    # return

    def write():
      while 1:
        data = f.read(256*1024)
        if not data: break
        conn.conn.write(data)
      f.close()

    return write

  def write_dir_listing(self, path):
    f = os.listdir(path)
    f = (os.path.isdir(os.path.join(path,i)) and i+'/' or i for i in f)

    r = ['<html>','<body>']
    r.extend('<a href="%s">%s</a><br />' % (i,i) for i in f)
    r.extend(['</html>','</body>'])
    return '\r\n'.join(r)



'''
About
=====
`pyy.web` contains useful classes and functions to aid in dealing with the HTTP
protocol and serving websites.


Usage
=====
Cookies
-------
`pyy.web.cookie` allows for easy creation of properly formatted cookies.

The cookie name and value are the only two required fields for its simplest
form:
    >>> cookie('test', 'cookie')
    <pyy.web.cookie.cookie test=cookie; path=/;>

You can also pass any number of the following arguments:
* `expires` - Expiration date (mutually exclusive with `duration`).
* `duration` - Length of cookie (mutually exclusive with `expires`).
* `path` - The URL path of the cookie (default: `/`).
* `domain` - The cookie's domain.
* `secure` - Whether or not the cookie is secure (default: `False`).
* `httponly` - If the cookie applies to only HTTP (default: `False`).

While the string and `__repr__` outputs contain the rendered cookie you can
also call `render()` directly which takes an optional `is_header` boolean.
    >>> cookie('test', 'cookie').render()
    'test=cookie; path=/;'
    >>> cookie('test', 'cookie').render(True)
    'Set-Cookie: test=cookie; path=/;'


HTTP Messages
-------------
`pyy.web.httpmessage` contains the classes for the `httprequest` and
`httpresponse` objects. These two classes hold information on the HTTP requests
and the HTTP responses, respectively, that the other modules pass between
themselves.


Parsers
-------
The `pyy.web.parsers` class contains four functions which aid in parsing common
formats of the HTTP protocol into a more friendly and usable format.

*   `parse_query(string)` - Parses `a=b&c=d` format into a dictionary. Also
    supports arrays.
*   `parse_semi(string)` - Parses `a=b; c=d` format into a dictionary.
*   `parse_user_agent(string)` - Parses a user agent string and returns a
    browser object containing vendor and version.
*   `parse_multipart(content_type, data)` - Parses multipart encoded form data
    into a dictionary.


Resolvers
---------
When serving dynamic web applications it is sometimes useful to obfuscate your
URLs or simple clean them up to be "pretty". The `resolvers` class allows you
to create logical rules or patterns for your URLs and have them easily be
mapped back to approprite classes.

There are currently two methods of dynamic URL resolving: Regex-based and file
heirarchy-based.

Regex-based resolving takes a list of tuples which associate a regular
expression to a class. When a request is resolved in this way the tuples are
iterated over and the first match is taken and returned.
    urls = [
      (r'^/photos/', pages.photos),
      (r'^/videos/', pages.videos),
    ]
If a match is not found, an `httperror` is raised with a 404 code.

You can also use named match groups in your regular expressions to help parse
the URL which will be copied into the request's `GET` mapping.
    urls = [
      (r'^/photos/((?P<id>\d+)/)?$', pages.photos)
    ]
Here an optional ID can be specified after the `/photos/` qualifier and the
`pages.photos` class can react accordingly.

--------
The other type of resolver uses a file system-based heirarchy to determine
which class to associate a URL with.

When you create a directory like the following:
    pages/
          __init__.py
          home.py
          admin/
                __init__.py
                users.py
you can use corresponding URLs to access those classes:
    /             (would match in __init__.py)
    /home/        (would match home.py)
    /home/page2/  (would match home.py with 'page2' put into the request's GET)
    /admin/       (would match admin/__init__.py)
    /admin/users/ (would match admin/users.py)
    /admin/other/ (would match admin/__init.py with 'other' put into the request's GET)

In each of those files should be an Index class which is the default one
selected in a file. If in the case of the `/admin/other/` URL the
`admin/__init__.py` file had an `Other` class it would be the one that was
selected.


Developed By
============
* Tom Flanagan - <theknio@gmail.com>
* Jake Wharton - <jakewharton@gmail.com>

Git repository located at
[github.com/Knio/pyy](http://github.com/Knio/pyy)


License
=======
    Copyright 2009 Tom Flanagan, Jake Wharton

    This file is part of pyy.

    pyy is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    pyy is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General
    Public License along with pyy.  If not, see
    <http://www.gnu.org/licenses/>.
'''

__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

from httpmessage import *
from page import page
from url import url



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''
import sys
from pyy.httpserver import *


from fileserver import fileserver
from pyyserver  import pyyscript

if __name__ == '__main__':
  args = list(sys.argv[1:])

  port = 8080
  dir  = '.'
  while args:
    c = args.pop(0)
    if c == '-p':
      port = int(args.pop(0))
    if c == '-d':
      dir = args.pop(0)


  class myserver(httpserver):
    port = port
    uri = [
      '^/(.*)$',                   fileserver(dir),
      '^/(.*(.py|c|h|js|java))$',  syntaxfileserver(dir),
      # ('^/(.*\.m)$',                gvimfileserver(dir)),
      # ('^/(.*\.pyy)$',              pyyscript(dir)),
    ]

  myserver().run()




__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import os
import sys
import imp
from urllib  import unquote_plus
from pyy.web import httperror


class pyyscript(object):
  def __init__(self, root, fname, *args, **kwargs):
    self.root   = root
    self.fname = os.path.join(self.root, unquote_plus(fname))
    self.mname = os.path.basename(fname).split('.')[0]
    self.dname = os.path.dirname(fname)
    self.mtime = 0
    self.args   = args
    self.kwargs = kwargs
    self.module = None


  def load_module(self):
    if not self.fname.startswith(self.root):
      raise httperror(403)

    if not os.path.exists(self.fname):
      raise httperror(404)

    if not os.path.isfile(self.fname):
      raise httperror(403)
    
    if os.path.getmtime(self.fname) == self.mtime:
      return self.module

    f = open(self.fname, 'U')
    sys.path.append(self.dname) # do we even need this?

    self.module = imp.load_module(self.mname, f, self.fname, ('.py', 'U', 1))
    self.mtime  = os.path.getmtime(self.fname)
    
    sys.path.remove(self.dname)
    return self.module

  def handle(self, handler, req, res, *args2, **kwargs2):
    m = self.load_module()
     
    h = getattr(m, req.method.lower(),
        getattr(m, 'handle', None))
    
    if not h:
      raise httperror(405) # method not allowed
    
    cwd = os.getcwd()
    os.chdir(self.root)
    args = self.args + args2
    kwargs = dict(self.kwargs)
    kwargs.update(kwargs2)
    r = h(handler, req, res, *args, **kwargs)
    os.chdir(cwd)
    return r

  def handle_error(self, handler, req, res, status, *args):
    m = self.load_module()
    
    h = getattr(m, 'handle_error', None)
    if not h: return
    return  h(handler, req, res, status, *args)






# cgi.py - pyy CGI gateway
#
# WARNING: Edit this file very carefully, a bug will crash your site!

__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import os
import sys
import imp


def get_request():
  env = os.environ
  req = httprequest()
  
  req.read   = sys.stdin.read
  req.method = env['REQUEST_METHOD']
  req.uri    = env['REQUEST_URI']
  
  http = env['SERVER_PROTOCOL']
  if http == 'HTTP/1.0':
    req.http = 1.0
  elif http == 'HTTP/1.1':
    req.http = 1.1

  HEADERS = {
    'CONTENT_LENGTH':       'Content-Length',
    'CONTENT_TYPE':         'Content-Type',
    'HTTP_USER_AGENT':      'User-Agent',
    'HTTP_COOKIE':          'Cookie',
    'HTTP_TE':              'Transfer-Encoding',
    'HTTP_COOKIE2':         'Cookie2',
    'HTTP_HOST':            'Host',
    'HTTP_CONNECTION':      'Connection',
    'HTTP_ACCEPT':          'Accept',
    'HTTP_ACCEPT_LANGUAGE': 'Accept-Language',
    'HTTP_ACCEPT_ENCODING': 'Accept-Encoding',
    'HTTP_ACCEPT_CHARSET':  'Accept-Charset',
  }
  
  for key, value in env.iteritems():
    if key in HEADERS:
      req.headers[HEADERS[key]] = value
  
  return req


def make_response(res):
  if res.body is None: res.body = ''
  res.headers.setdefault('Content-Type', 'text/plain; charset=ISO-8859-4')
  res.headers.setdefault('Cache-Control', 'no-cache')
  #res.headers.setdefault('Content-Length', len(res.body))

def write_response(res):
  if res.status:
    print 'Status: %d %s' % (res.statusnum, res.statusmsg)
  
  for key, value in res.headers.iteritems():
    print '%s: %s' % (key, value)

  print
  print res.body



if 'REQUEST_METHOD' in os.environ:
  import os
  import sys
  import imp
  from pyy.web import *
  
  fname = sys.argv[0]
  mname = os.path.basename(fname).split('.')[0]
  dname = os.path.dirname(fname)
  sys.path.append(dname)
  
  f = open(fname, 'U')
  
  try:
    m   = imp.load_module(mname, f, fname, ('.py', 'U', 1))
    req = get_request()
    res = httpresponse()
    
    result = None
    try:
      h = getattr(m, req.method.lower(), getattr(m, 'handle', None))
      if not h:
        raise httperror(405) # method not allowed
      result = h(None, req, res)
      
    except httperror, e:
      res = httpresponse()
      res.status = e.args[0]
      res.body   = '%d %s' % (res.statusnum, res.statusmsg)
      # TODO let the page do custom error handling

    make_response(res)
    write_response(res)
    if result:
      result()

  except SystemExit:
    pass

  except Exception, e:
    import traceback
    print 'Content-Type: text/html'
    print
    print '<html><head><title>Error</title></head><body><pre>'
    print 'An error occured while generating this page:'
    print
    print traceback.format_exc()
    print '</pre></body></html>'
      
  f.close()
  sys.exit(0)



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

import numbers
from collections import defaultdict, namedtuple
from functools import wraps
import threading


def _get_thread_context():
  context = [threading.current_thread()]
  try:
    import greenlet
    context.append(greenlet.getcurrent())
  except:
    pass
  return hash(tuple(context))


class pyy_tag(object):
  TAB = '  '  # TODO make this a parameter to render(), and a tag.

  is_single = False  # Tag does not require matching end tag (ex. <hr/>)
  is_pretty = True   # Text inside the tag should be left as-is (ex. <pre>)
                     # otherwise, text will be escaped() and whitespace may be
                     # modified

  frame = namedtuple('frame', ['tag', 'items', 'used'])

  def __new__(_cls, *args, **kwargs):
    '''
    Check if bare tag is being used a a decorator.
    decorate the function and return
    '''
    if len(args) == 1 and callable(args[0]) \
        and not isinstance(args[0], pyy_tag) and not kwargs:
      wrapped = args[0]

      @wraps(wrapped)
      def f(*args, **kwargs):
        with _cls() as _tag:
          return wrapped(*args, **kwargs) or _tag
      return f
    return object.__new__(_cls)

  def __init__(self, *args, **kwargs):
    '''
    Creates a new tag. Child tags should be passed as aruguments and attributes
    should be passed as keyword arguments.

    There is a non-rendering attribute which controls how the tag renders:

    * `__inline` - Boolean value. If True renders all children tags on the same
                   line.
    '''

    self.attributes = {}
    self.children   = []
    self.parent     = None
    self.document   = None

    # Does not insert newlines on all children if True (recursive attribute)
    self.do_inline = kwargs.pop('__inline', False)

    #Add child elements
    self.add(*args)

    for attr, value in kwargs.items():
      self.set_attribute(*pyy_tag.clean_pair(attr, value))

    ctx = pyy_tag._with_contexts[_get_thread_context()]
    if ctx and ctx[-1]:
      ctx[-1].items.append(self)

  # stack of (root_tag, [new_tags], set(used_tags))
  _with_contexts = defaultdict(list)

  def __enter__(self):
    ctx = pyy_tag._with_contexts[_get_thread_context()]
    ctx.append(pyy_tag.frame(self, [], set()))
    return self

  def __exit__(self, type, value, traceback):
    ctx = pyy_tag._with_contexts[_get_thread_context()]
    slf, items, used = ctx[-1]
    ctx[-1] = None
    for item in items:
      if item in used: continue
      self.add(item)
    ctx.pop()

  def __call__(self, func):
    '''
    tag instance is being used as a decorator.
    wrap func to make a copy of this tag
    '''
    @wraps(func)
    def f(*args, **kwargs):
      with type(self)() as _tag:
        _tag.children = self.children[:]
        _tag.attributes = dict(self.attributes)
        return func(*args, **kwargs) or _tag
    return f

  def set_attribute(self, key, value):
    '''
    Add or update the value of an attribute.
    '''
    if isinstance(key, int):
      self.children[key] = value
    elif isinstance(key, basestring):
      self.attributes[key] = value
    else:
      raise TypeError('Only integer and string types are valid for assigning '
          'child tags and attributes, respectively.')
  __setitem__ = set_attribute

  def setdocument(self, doc):
    '''
    Creates a reference to the parent document to allow for partial-tree
    validation.
    '''
    # assume that a document is correct in the subtree
    if self.document != doc:
      self.document = doc
      for i in self.children:
        if not isinstance(i, pyy_tag): return
        i.setdocument(doc)

  def add(self, *args):
    '''
    Add new child tags.
    '''
    for obj in args:
      if isinstance(obj, numbers.Number):
        # Convert to string so we fall into next if block
        obj = str(obj)

      if isinstance(obj, basestring):
        if self.is_pretty:
          obj = escape(obj)
        self.children.append(obj)

      elif isinstance(obj, pyy_tag):
        ctx = pyy_tag._with_contexts[_get_thread_context()]
        if ctx and ctx[-1]:
          ctx[-1].used.add(obj)
        self.children.append(obj)
        obj.parent = self
        obj.setdocument(self.document)

      elif isinstance(obj, dict):
        for attr, value in obj.items():
          self.set_attribute(*pyy_tag.clean_pair(attr, value))

      elif hasattr(obj, '__iter__'):
        for subobj in obj:
          self.add(subobj)

      else:  # wtf is it?
        raise ValueError('%r not a tag or string.' % obj)

    if len(args) == 1:
      return args[0]

    return args

  def add_raw_string(self, s):
    self.children.append(s)

  def clear(self):
    for i in self.children:
      if isinstance(i, pyy_tag) and i.parent is self:
        i.parent = None
    self.children = []

  def get(self, tag=None, **kwargs):
    '''
    Recursively searches children for tags of a certain
    type with matching attributes.
    '''
    # Stupid workaround since we can not use pyy_tag in the method declaration
    if tag is None: tag = pyy_tag

    attrs = [(pyy_tag.clean_attribute(attr), value)
        for attr, value in kwargs.items()]

    results = []
    for child in self.children:
      if (isinstance(tag, basestring) and type(child).__name__ == tag) or \
        (not isinstance(tag, basestring) and isinstance(child, tag)):

        if all(child.attributes.get(attribute) == value
            for attribute, value in attrs):
          # If the child is of correct type and has all attributes and values
          # in kwargs add as a result
          results.append(child)
      if isinstance(child, pyy_tag):
        # If the child is a pyy_tag extend the search down through its children
        results.extend(child.get(tag, **kwargs))
    return results

  def __getitem__(self, key):
    '''
    Returns the stored value of the specified attribute or child
    (if it exists).
    '''
    if isinstance(key, int):
      # Children are accessed using integers
      try:
        return object.__getattribute__(self, 'children')[key]
      except KeyError:
        raise IndexError('Child with index "%s" does not exist.' % key)
    elif isinstance(key, basestring):
      # Attributes are accessed using strings
      try:
        return object.__getattribute__(self, 'attributes')[key]
      except KeyError:
        raise AttributeError('Attribute "%s" does not exist.' % key)
    else:
      raise TypeError('Only integer and string types are valid for accessing '
          'child tags and attributes, respectively.')
  __getattr__ = __getitem__

  def __len__(self):
    '''
    Number of child elements.
    '''
    return len(self.children)

  def __nonzero__(self):
    '''
    Hack for "if x" and __len__
    '''
    return True

  def __iter__(self):
    '''
    Iterates over child elements.
    '''
    return self.children.__iter__()

  def __contains__(self, item):
    '''
    Checks recursively if item is in children tree.
    Accepts both a string and a class.
    '''
    return bool(self.get(item))

  def __iadd__(self, obj):
    '''
    Reflexive binary addition simply adds tag as a child.
    '''
    self.add(obj)
    return self

  def render(self, indent=1, inline=False):
    '''
    Returns a well-formatted string representation of the tag and renderings
    of all its child tags.
    '''
    inline = self.do_inline or inline

    t = type(self)
    name = getattr(t, 'tagname', t.__name__)

    # Workaround for python keywords and standard classes/methods
    # (del, object, input)
    if name[-1] == '_':
      name = name[:-1]

    rendered = ['<', name]

    for attribute, value in self.attributes.items():
      rendered.append(' %s="%s"' % (attribute, escape(unicode(value), True)))

    rendered.append('>')

    if not self.is_single:
      rendered.append(self._render_children(indent, inline))

      # if there are no children, or only 1 child that is not an html element,
      # do not add tabs and newlines
      no_children = self.is_pretty and self.children and \
          (not (len(self.children) == 1 and not isinstance(self[0], pyy_tag)))

      if no_children and not inline:
        rendered.append('\n')
        rendered.append(pyy_tag.TAB * (indent - 1))

      rendered.append('</')
      rendered.append(name)
      rendered.append('>')

    return u''.join(rendered)

  # String and unicode representations are the same as render()
  def __unicode__(self):
    return self.render()
  __str__ = __unicode__

  def _render_children(self, indent=1, inline=False):
    children = []
    for child in self.children:
      if isinstance(child, pyy_tag):
        if not inline and self.is_pretty:
          children.append('\n')
          children.append(pyy_tag.TAB * indent)
        children.append(child.render(indent + 1, inline))
      else:
        children.append(unicode(child))
    return ''.join(children)

  def __repr__(self):
    name = '%s.%s' % (self.__module__, type(self).__name__)

    attributes_len = len(self.attributes)
    attributes = '%s attribute' % attributes_len
    if attributes_len != 1: attributes += 's'

    children_len = len(self.children)
    children = '%s child' % children_len
    if children_len != 1: children += 'ren'

    return '<%s at %x: %s, %s>' % (name, id(self), attributes, children)

  @staticmethod
  def clean_attribute(attribute):
    '''
    Since some attributes match python keywords we prepend them with
    underscores. Python also does not support colons in keywords so underscores
    mid-attribute are replaced with colons.
    '''
    # shorthand
    attribute = {
      'cls': 'class',
    }.get(attribute, attribute)

    # Workaround for python's reserved words
    if attribute[0] == '_': attribute = attribute[1:]

    # Workaround for inability to use colon in python keywords
    if attribute in set(['http_equiv']) or attribute.startswith('data_'):
      return attribute.replace('_', '-').lower()
    return attribute.replace('_', ':').lower()

  @staticmethod
  def clean_pair(attribute, value):
    '''
    This will call `clean_attribute` on the attribute and also allows for the
    creation of boolean attributes.

    Ex. input(selected=True) is equivalent to input(selected="selected")
    '''
    attribute = pyy_tag.clean_attribute(attribute)

    # Check for boolean attributes
    # (i.e. selected=True becomes selected="selected")
    if value is True:
      value = attribute

    if value is False:
      value = "false"

    return (attribute, value)


def attr(*args, **kwargs):
  '''
  Set attributes on the current active tag context
  '''
  ctx = pyy_tag._with_contexts[_get_thread_context()]
  if ctx and ctx[-1]:
    dicts = args + (kwargs,)
    for d in dicts:
      for attr, value in d.items():
        ctx[-1].tag.set_attribute(*pyy_tag.clean_pair(attr, value))
  else:
    raise ValueError('not in a tag context')


# escape() is used in render
from util import escape



import re
import os
import imp

'''
Recursive routing utility

A config object is a list of (key, handler) tuples

If multiple handlers would match, the last one is chosen

A key can be:

  1) A regex
      The key matches if the regex matches the URL given

  2) An httperror instance



A handler object can be:

  1) An array
      The part of the URL that matches the regex will be stripped off,
      and the router will search recursively on the rest using the array
      as the new config object



'''


route = [
  ('^/')

]

def route(config, url):
  pass




__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy.  If not, see
<http://www.gnu.org/licenses/>.
'''

# http://guide.python-distribute.org/creation.html
from setuptools import setup, find_packages

version = '1.1.13'


setup(
  name    = 'pyy',
  version = version,
  author  = 'Tom Flanagan and Jake Wharton',
  author_email = 'theknio@gmail.com',
  license = 'LICENSE.txt',

  url          = 'http://github.com/Knio/pyy/',
  # download_url = 'http://github.com/Knio/pyy/tarball/' + version,

  description      = 'Python library for creating (X)HTML pages with the use of objects.',
  long_description = 'Python library for creating (X)HTML pages with the use of objects. This allows you to tightly integrate (X)HTML generation into your backend without the need of using an intermediate templating language.',
  keywords         = 'framework templating template html xhtml python html5',

  classifiers = [
    'Programming Language :: Python',
    'Operating System :: OS Independent',
    'Topic :: Software Development :: Libraries :: Application Frameworks',
    'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    'Intended Audience :: Developers',
  ],
  packages   = find_packages(),
  install_requires = [],
  include_package_data = True,
)



import re
from collections import defaultdict

URL_REGEX = ('^('
    '((?P<protocol>\w+)://)?'
    '(?P<host>[^\\/\\\\?]+)?'
    # TODO add ports
    '(?P<path>[^?]+)?'
    '(\?(?P<args>[^#]*))?'
    '(#(?P<anchor>.+))?'
    ')$')


FORM_URLENCODING_REGEX = ('^('
    '(?P<var>[^=]+=[^&]*(&[^=]+=[^&]*)*)?'
    ')$')


class url(object):
    def __init__(self, data=None, **kwargs):
        self.protocol = None
        self.host = None
        self.path = '/'
        self.args = {}
        self.anchor = None
        if data is None:
            pass
        elif isinstance(data, basestring):
            m = re.match(URL_REGEX, data)
            if m:
                self.update(**dict(i for i in m.groupdict().items() if i[1]))
        elif isinstance(data, url):
            self.protocol = data.protocol
            self.host = data.host
            self.path = data.path
            self.args = data.args
            self.anchor = data.anchor
            if isinstance(self.args, dict):
                self.args = dict(self.args)
        else:
            raise ValueError

        self.update(**kwargs)

        a = self.parse_arguments(self.args)
        if a:
            self.args = a

    def update(self, **args):
        # TODO check args are only proto, host, path, etc
        for k, v in args.iteritems():
            setattr(self, k, v)
        return self

    def update_args(self, *a, **kw):
        # TODO whaf if args isnt a dict?
        self.args.update(*a, **kw)
        return self

    @staticmethod
    def parse_arguments(data):
        if not isinstance(data, basestring):
            return None
        match = re.match(FORM_URLENCODING_REGEX, data)
        if not match:
            return None
        argstring = match.group('var')
        if not argstring:
            return None

        args = defaultdict(list)
        for j in argstring.split('&'):
            k, x = j.split('=', 1)
            args[k].append(x)

        argsf = {}
        for k, v in args.items():
            if len(v) == 1:
                argsf[k] = v[0]
            else:
                argsf[k] = v

        return argsf

    def get_arg(self, arg, default=None):
        return self.args and self.args.get(arg, default)

    def get_args(self, arg):
        if not isinstance(self.args, dict):
            return None
        v = self.args.get(arg, [])
        if not isinstance(v, (list, tuple)):
            return [v]
        return v

    __getitem__ = get_arg

    def __setitem__(self, arg, val):
        if not isinstance(self.args, dict):
            raise ValueError('url arguments are not form encoded')
        self.args[arg] = val

    def __str__(self):
        ret = []
        host = self.host
        if self.protocol is not None:
            ret.append('%s://' % self.protocol)
            host = host or ''

        if host is not None:
            ret.append(host)

        if self.path is not None:
            ret.append(unicode(self.path))

        if self.args is None:
            pass
        elif isinstance(self.args, basestring):
            ret.append('?%s' % self.args)
        elif isinstance(self.args, dict):
            a = []
            for k, v in self.args.iteritems():
                if isinstance(v, (basestring, int, long, float)):
                    a.append('%s=%s' % (unicode(k), unicode(v)))
                if isinstance(v, (list, tuple)):
                    for j in v:
                        a.append('%s=%s' % (unicode(k), unicode(j)))
            if a:
                ret.append('?' + '&'.join(a))
        else:
            ret.append('?' + unicode(self.args))

        if self.anchor is not None:
            ret.append('#' + unicode(self.anchor))

        return ''.join(ret)

    __unicode__ = __str__


if __name__ == '__main__':
    print url()
    print url('/moo')
    print url('http://moo')
    print url('http:///moo')
    print url('http://moo.com')
    print url('http://moo.com/')
    print url('http://moo.com/foo')
    print url('http://moo.com/foo/bar/')
    print url('http://moo.com/foo/bar?')
    print url('http://moo.com/foo/bar/?var=1')
    print url('http://moo.com/foo/bar/?var=1&baz=dog')
    print url('http://moo.com/foo/bar/?var=1&baz=dog&var=3')
    print url('http://moo.com/foo/bar/?var=1&baz=dog&var=3#here')
    assert url('?moo=1')['moo'] == '1'
    print url(url('?moo=1'))
    print url(url('?moo=1')).update_args(bar=2)



__license__ = '''
This file is part of pyy.

pyy is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

pyy is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General
Public License along with pyy. If not, see
<http://www.gnu.org/licenses/>.
'''

from pyy.html.document import document
import a

class Index(document): pass
